from marshmallow import Schema
from .utils import make_schema, NonceSchema

try:
    from apispec import APISpec, utils
    from apispec.ext.marshmallow import MarshmallowPlugin
    from apispec.ext.marshmallow.common import resolve_schema_cls
    import yaml
except ImportError:
    raise ImportError(
        'You must have the `apispec` package installed to use this feature.  '
        'You can install it with `pip install pyramid_marshmallow[openapi].'
    )


def schema_name_resolver(schema):
    # TODO:  In apispec 3.0.0, this function will receive a schema instance.
    # Should introspect instance to handle schema modifiers, maybe treating it
    # as nonce schema.
    schema_cls = resolve_schema_cls(schema)
    name = schema_cls.__name__
    if issubclass(schema_cls, NonceSchema):
        return None
    if name.endswith("Schema"):
        return name[:-6] or name
    return name


def list_paths(introspector):
    for item in introspector.get_category('views'):
        spect = item['introspectable']
        path = make_path(introspector, spect)
        if path is None:
            continue
        operations = dict()
        methods = spect['request_methods'] or ['GET']
        if isinstance(methods, str):
            methods = [methods]
        for method in methods:
            method = method.lower()
            operations[method] = spect
        yield path, operations


def make_path(introspector, introspectable):
    if introspectable['route_name']:
        route = introspector.get('routes', introspectable['route_name'])
        return route['pattern']
    elif introspectable['context']:
        context = introspectable['context']
        base = getattr(context, '__path__', None)
        if base is None:
            return None
        else:
            return base + '/' + (introspectable['name'] or '')
    else:
        return None


def _schema(schema):
    if isinstance(schema, dict):
        return make_schema(schema)
    else:
        return schema


def split_docstring(docstring):
    """
    Split a docstring in half, delineated with a "---".  The first half is
    returned verbatim, the second half is parsed as YAML.

    """
    split_lines = utils.trim_docstring(docstring).split('\n')

    # Cut YAML from rest of docstring
    for index, line in enumerate(split_lines):
        line = line.strip()
        if line.startswith('---'):
            cut_from = index
            break
    else:
        return (docstring or '').strip(), dict()

    docs = '\n'.join(split_lines[:cut_from]).strip()
    yaml_string = '\n'.join(split_lines[cut_from:])
    yaml_string = utils.dedent(yaml_string)
    parsed = yaml.load(yaml_string)
    return docs, parsed


def set_request_body(spec, op, view):
    op['requestBody'] = {
        'content': {
            'application/json': {
                'schema': _schema(view['validate']),
            },
        },
    }


def set_query_params(spec, op, view):
    op['parameters'].append({
        'in': 'query',
        'schema': _schema(view['validate']),
    })


def set_response_body(spec, op, view):
    op['responses']['200'] = {
        'description': '',
        'content': {
            'application/json': {
                'schema': _schema(view['marshal']),
            },
        },
    }


def set_url_params(spec, op, view):
    context = view['context']
    if not context:
        return
    params = getattr(context, '__params__', [])
    for param in params:
        param['in'] = 'path'
        param['required'] = True
    op['parameters'].extend(params)


def set_tag(spec, op, view):
    context = view['context']
    if not context:
        return
    tag = getattr(context, '__tag__', None)
    if not tag:
        return
    if isinstance(tag, dict):
        # Cheating and using the private variable spec._tags
        if not any(x['name'] == tag['name'] for x in spec._tags):
            spec.tag(tag)
        tag_name = tag['name']
    else:
        tag_name = tag
    op.setdefault('tags', []).append(tag_name)


def create_spec(title, version, introspector):
    marshmallow_plugin = MarshmallowPlugin(
        schema_name_resolver=schema_name_resolver,
    )
    spec = APISpec(
        title=title,
        version=version,
        openapi_version='3.0.2',
        plugins=[marshmallow_plugin],
    )
    for path, operations in list_paths(introspector):
        final_ops = dict()
        for method, view in operations.items():
            docstring, op = split_docstring(view['callable'].__doc__)
            op.setdefault('responses', dict())
            op.setdefault('description', docstring)
            op.setdefault('summary', path)
            op.setdefault('parameters', [])
            set_url_params(spec, op, view)
            if 'validate' in view and method != 'get':
                set_request_body(spec, op, view)
            elif 'validate' in view:
                set_query_params(spec, op, view)
            if 'marshal' in view:
                set_response_body(spec, op, view)
            set_tag(spec, op, view)
            # We are required to have some response, so make one up.
            if not op['responses']:
                op['responses']['200'] = {
                    'description': '',
                }
            final_ops[method] = op
        spec.path(path, operations=final_ops)

    return spec



from marshmallow import Schema, ValidationError
from pyramid.viewderivers import VIEW

from .utils import make_schema, NonceSchema


__all__ = [
    'make_schema',
    'NonceSchema',
]


def includeme(config):
    config.add_view_deriver(view_validator)
    config.add_view_deriver(view_marshaller, under='rendered_view', over=VIEW)


def process_schema(schema):
    """
    Handle a schema passed in as a view deriver, creating a nonce schema if a
    dictionary.

    """
    if schema is None:
        return None
    elif isinstance(schema, Schema):
        return schema
    elif isinstance(schema, dict):
        _Schema = make_schema(schema)
        return _Schema()
    else:
        raise TypeError('Schema is of invalid type.')


def view_validator(view, info):
    schema = process_schema(info.options.get('validate'))
    if schema is None:
        return view

    def wrapped(context, request):
        if request.method == 'GET':
            data = dict()
            for k, v in request.GET.items():
                data[k] = v
        else:
            data = request.json_body
        request.data = schema.load(data)
        return view(context, request)

    return wrapped


view_validator.options = ('validate',)


def view_marshaller(view, info):
    schema = process_schema(info.options.get('marshal'))
    if schema is None:
        return view

    def wrapped(context, request):
        output = view(context, request)
        return schema.dump(output)

    return wrapped


view_marshaller.options = ('marshal',)



from io import BytesIO
import zipfile
import shutil
import pkg_resources
import argparse
import json
from pyramid.paster import get_app


parser = argparse.ArgumentParser()
parser.add_argument(
    'ini',
    help='The .ini config file for the Pyramid project.',
)
parser.add_argument(
    '--format',
    help='The output, one of "json", "yaml", or "html".',
    default='json',
)
parser.add_argument(
    '--output',
    help='The file to output to.',
    default='-',
)


def generate():
    from .spec import create_spec
    args = parser.parse_args()
    app = get_app(args.ini)
    settings = app.registry.settings
    title = settings.get('openapi.title', 'Untitled')
    version = settings.get('openapi.version', '0.0.0')
    introspector = app.registry.introspector
    spec = create_spec(title, version, introspector)
    if args.format == 'json':
        output = json.dumps(spec.to_dict())
    elif args.format == 'yaml':
        output = spec.to_yaml()
    elif args.format == 'html':
        output = generate_html(spec)
    else:
        raise ValueError('Format must be one of "json", "yaml", or "html".')
    if args.output == '-':
        print(output)
    else:
        with open(args.output, 'wb') as fh:
            if isinstance(output, str):
                output = output.encode('utf8')
            fh.write(output)


def generate_html(spec):
    data = json.dumps(spec.to_dict())
    return HTML_TEMPLATE.format(
        title=spec.title,
        version=spec.version,
        spec=data,
    )


HTML_TEMPLATE = """
<!DOCTYPE html>
<html>
    <head>
        <title>{title} {version}</title>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
        <style>
            body {{
                margin: 0;
                padding: 0;
            }}
        </style>
    </head>
    <body>
        <div id="redoc"></div>
        <script type="text/json" id="spec">{spec}</script>
        <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"></script>
        <script>
            window.addEventListener('load', function() {{
                var el = document.getElementById('redoc');
                var spec = JSON.parse(document.getElementById('spec').text);
                Redoc.init(spec, {{}}, el);
            }});
        </script>
    </body>
</html>
"""


if __name__ == '__main__':
    generate()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pyramid_marshmallow']

package_data = \
{'': ['*']}

install_requires = \
['marshmallow>=3.0,<4.0', 'pyramid>=1.7,<2.0']

extras_require = \
{'openapi': ['apispec>=2.0,<3.0', 'PyYAML>=3.10']}

entry_points = \
{'console_scripts': ['generate-spec = '
                     'pyramid_marshmallow.generatespec:generate']}

setup_kwargs = {
    'name': 'pyramid-marshmallow',
    'version': '0.5.0',
    'description': 'Validate request and response data with Marshmallow and optionally generate an OpenAPI spec.',
    'long_description': '# pyramid-marshmallow\n\npyramid-marshmallow is a simple Pyramid plugin that allows you to validate and\nmarshal a JSON HTTP request or response using\n[Marshmallow](http://marshmallow.readthedocs.io/) schemas.  You can then\nleverage this to automatically generate an OpenAPI specification for your API.\n\n## Basic usage\n\nInstall the project with `pip install pyramid-marshmallow`.\n\nActivate it by adding `config.include(\'pyramid_marshmallow\')` into your config\nfunction or adding `pyramid.includes = pyramid_marshmallow` into your ini file.\n\nTo validate incoming data, set `validate` to a Marshmallow schema in your\n`view_config`.  The request body is parsed as JSON then passed through the\nschema\'s `load` function.  You can access the processed data with\n`request.data`.\n\n```python\nfrom marshmallow import Schema, String\n\n\nclass HelloSchema(Schema):\n    name = String()\n\n\n@view_config(\n    context=Root,\n    name=\'hello\',\n    request_method=\'post\',\n    validate=HelloSchema(),\n)\ndef hello(context, request):\n    return Response(body=\'Hello, {}\'.format(\n        request.data[\'name\']\n    ))\n```\n\nFor GET requests, the URL parameters are passed into the schema.  Value lists\nare not currently supported.\n\nSetting `marshal` in your `view_config` will run the view output through\nmarshmallow (i.e. `Schema.dump`) before going to the renderer.  You probably\nwill want to set the renderer to `json`.\n\n```\n@view_config(\n    context=Root,\n    name=\'hello\',\n    request_method=\'get\',\n    marshal=HelloSchema(),\n    renderer=\'json\',\n)\ndef hello(context, request):\n    name = fetch_name()\n    return {\n        \'name\': name,\n    }\n```\n\n`validate` and `marshal` operate independently, so can be used separately or\ntogether.\n\nAs a convenience, you can pass in a dictionary to `validate` or `marshal` and\npyramid-marshmallow will turn it into a schema for you.\n\n```python\n@view_config(\n    context=Root,\n    name=\'hello\',\n    request_method=\'post\',\n    validate={\n        \'name\': String(),\n    },\n)\n```\n\nYou can also get a schema made from a dictionary by using the\n`pyramid_marshmallow.make_schema` function.  This can be useful for `Nested`\nfields.\n\n\n### Error handling\n\npyramid-marshmallow passes through exceptions from marshmallow.  So errors\nduring validation will raise a `marshmallow.exceptions.ValidationError`\nexception.\n([Documentation](https://marshmallow.readthedocs.io/en/stable/api_reference.html#marshmallow.exceptions.ValidationError))\n\nYou may want to attach a view to this exception to expose the error messages to\nthe user.\n\n```python\n@view_config(\n    context=ValidationError,\n    renderer=\'json\',\n)\ndef validation_error(context, request):\n    request.response.status = 401  # HTTP Bad Request\n    return {\n        \'errors\': context.normalized_messages(),\n    }\n```\n\n## OpenAPI\n\nBy adding validation and marshalling to your views, we have the opportunity to\nutilize that data to generate documentation.  pyramid-marshmallow includes an\nutility that uses [apispec](https://apispec.readthedocs.io/en/stable/) to\ngenerate an [OpenAPI](https://swagger.io/resources/open-api/) specification for\nyour application.\n\nFirst, you\'ll need to install some extra dependencies.\n\n```bash\npip install pyramid-marshmallow[openapi]\n```\n\nNow you can generate your spec by simply passing in an ini file.\npyramid-marshmallow needs to run your application in order to inspect it, so\nthe ini file should contain all the necessary configuration to do so.\n\n```bash\ngenerate-spec development.ini\n```\n\nThis will output the spec to stdout as JSON.  You can set the `--output` flag\nto output the results to a file.\n\nYou can set `--format yaml` to output the spec as YAML instead or\n`--format html` to output the spec as an HTML file, powered by\n[ReDoc](https://github.com/Redocly/redoc).\n\nBy default, your spec will be titled "Untitled" and versioned "0.1.0".  You can\nchange this by setting `openapi.title` and `openapi.version` in your ini file.\n\n### Additional Documentation\n\nTo add additional documentation to schema fields, you can set the `description`\nproperty.\n\n```python\nclass Hello(Schema):\n    name = String(required=True, description=\'Your first and last name.\')\n```\n\nDocumentation for the endpoint will be pulled from the view callable\'s\ndocstring.\n\nYou can also augment the spec by adding a line of three hyphens followed by\nYAML.  The YAML will be parsed and merged into to the endpoint\'s spec.  This\ncan be useful for documenting endpoints that cannot be validated or marshalled,\nsuch as endpoints that return an empty body.\n\n```python\n@view_config(\n    context=WidgetResource,\n    method=\'post\',\n    validate=WidgetSchema(),\n)\ndef create_widget(context, request):\n    """\n    Create a new widget.\n    ---\n    responses:\n        201:\n            description: Indicates the widget was successfully created.\n    """\n    create_widget()\n    return HTTPCreated()\n```\n\n## URL Traversal\n\nIf you\'re using Pyramid\'s URL traversal, the generated spec may be mostly\nempty.  This is because pyramid-marshmallow has no way of knowing where in the\nresource tree a resource is.  You can denote this by setting the `__path__`\nproperty on each resource.\n\n```python\nclass Widget(Resource):\n    __path__ = \'/widget\'\n```\n\nViews attached to this resource will then be added to the spec.\n\nYou can add parameters to your path via the `__params__` property.  You can\nalso tag all attached views via `__tag__`.  Once you define a tag in one\nresource, you can use it elsewhere by setting `__tag__` to the tag name.\n\n```python\nclass Widget(Resource):\n    __path__ = \'/widget/{widgetId}\'\n    __params__ = [{\n        \'name\': \'widgetId\',\n        \'schema\': {\n            \'type\': \'integer\',\n        },\n    }]\n    __tag__ = {\n        \'name\': \'widgets\',\n        \'description\': \'Endpoints for managing a widget.\',\n    }\n```\n\n## Prior Art\n\n[pyramid-apispec](https://pypi.org/project/pyramid-apispec/) allows you to\naugment view callable docstrings with OpenAPI definitions and can reference\nMarshmallow schemas with the apispec Marshmallow plugin.  It does not support\nvalidating input and marshalling output.  Schemas and routes must be manually\ndeclared.\n\n[Cornice](https://cornice.readthedocs.io/en/latest/schema.html#using-marshmallow)\nsupports validation with Marshmallow schemas, however only on Cornice\nresources, not arbitrary Pyramid endpoints.  It does not support\nauto-generating OpenAPI documentation.\n',
    'author': 'Theron Luhn',
    'author_email': 'theron@luhn.com',
    'url': 'https://github.com/luhn/pyramid-marshmallow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



from marshmallow import Schema


class NonceSchema(Schema):
    """
    Base class for nonce schemas (schemas created on-the-fly from
    dictionaries).

    """
    pass


def make_schema(schema=None, **kwargs):
    """
    Create a schema from a dictionary.

    """
    if schema is None:
        schema = kwargs
    return type('_Schema', (NonceSchema,), schema.copy())
